//
// Server for Miss Connectivity
//
// TCP/IP Testprogram
// Stephan Diehl (diehl@cs.uni-sb.de), March 1998

// Java APIs
import java.util.*;
import java.net.*;
import java.io.*;

// The StatueServer starts three threads:
// 1) the StatueServer itself: it listens for new connections and
//    if a client tries to connect it starts a new thread (2) for
//    this connection. It also starts a thread (3) to detect and
//    remove broken connections.
// 2) ClientEntry: Every client entry is also a thread. It receives
//    all messages from and sends all messages to that client. There
//    are two kinds of messages: ALIVE messages and updates. Updates
//    received from one client are send to all others.
// 3) StillAliveThread: It waits for several seconds. Then it checks
//    for every connection, what amount of time passed since the
//    last ALIVE message. If this amount is higher then a certain
//    threshold it closes the connection and removes its ClientEntry.
 

public class StatueServer extends Thread
{ ServerSocket acceptsock;
  Socket newsock;
  int port = 3000;
  static Vector clients = new Vector();
 
  static int partnum=12;
  static String[] geos = new String[partnum];
  static String[] apps = new String[partnum];

  static FileWriter log;
  static String logfilename="session.log";

  static void errout(String str) { System.out.println(str); }

  public static void main(String argv[])
  { Thread t=new StatueServer();
    for(int i=0;i<partnum;i++) { geos[i]="GEO1"; apps[i]="APP1"; }
    t.start();
    Thread.yield();
  }

  public void run()
  { StillAliveThread checkThread = new StillAliveThread(this);
    try {log=new FileWriter(logfilename,true);}
    catch (IOException e) 
      { errout("Could not open log file "+logfilename); }
    
    try { acceptsock=new ServerSocket(port); }
    catch (IOException e)
      { errout("Unable to open socket and start server!"); return; }
    while(true)
      { try { newsock = acceptsock.accept(); }
        catch (IOException e)
	  { errout("Unable to accept connections!"); 
            newsock=null;
          }
        clients.addElement(new ClientEntry(newsock));
        Thread.yield();
      }
  }

static void sendUpdates(ClientEntry sender, String str)
  { int i;
    if (str.indexOf("TLK")==0) 
     { if (log!=null)
        { try { log.write(str+"\n"); log.flush(); }
          catch (IOException e) 
            { StatueServer.errout("Could not write to log file "+logfilename); }
        }
     }
    synchronized (clients)
    { for(i=0; i<clients.size();i++)
       { ClientEntry cl=(ClientEntry) clients.elementAt(i);
         if (cl!=sender) cl.sendUpdate(str);
       }
    }
  }

} // end of class StatueServer

// -----------------------------------

class ClientEntry extends Thread
{  int port = 3001;         // port for communication with client
   Socket tcpsock;
   BufferedReader tcpinput;
   PrintWriter tcpoutput;
   public long lastAlive;
   public static int maxWaitForClient=20000;

 void errout(String str) { System.out.println(str); }

 public ClientEntry(Socket newsock) 
  { tcpsock=newsock;
    errout("Connecting to Client "
           + tcpsock.getInetAddress().getHostName());    
    connect();
    this.start();
  }

 public  void connect() // connect to client and send the latest update to it
  { try { tcpinput = new BufferedReader(new InputStreamReader(tcpsock.getInputStream())); }
    catch (IOException e) { errout("Could not open input stream from client!"); return; }
    try { tcpoutput = new PrintWriter(tcpsock.getOutputStream(),true); }
    catch (IOException e) { errout("Could not open output stream to client!"); return; }
    String update="UPD ";
    for (int i=0; i<StatueServer.partnum; i++) 
      { update=update+StatueServer.geos[i]+";"+StatueServer.apps[i]+";"; }
    this.sendUpdate(update);
  }

 public void disconnect()
  { this.stop();
    try { tcpinput.close();
          tcpoutput.close();
          tcpsock.close();
        }
  catch (IOException ex) 
    { errout("Could not close connection from "+tcpsock.getInetAddress().getHostName()); }
  }

 public void run()
  { boolean alive=true;
    lastAlive=System.currentTimeMillis();
    String str; 
    while (alive)
    { str=this.getUpdate();
      if (str==null) { Thread.yield(); continue; }
      if (str.equals("ALIVE")) 
         { lastAlive=System.currentTimeMillis();
         }
      if (str.indexOf("TLK")==0) 
	 { StatueServer.sendUpdates(null,str); // null: send it also back to the user
         } 
      if (str.indexOf("CHG")==0) 
         { int parti;
           int cur=4;
           StatueServer.sendUpdates(this,str);
           parti=Integer.parseInt(str.substring(cur,str.indexOf(':')));
           cur=str.indexOf(':')+1;
           StatueServer.geos[parti]=str.substring(cur,cur+4); 
           cur=str.indexOf(';',cur)+1;
           StatueServer.apps[parti]=str.substring(cur,cur+4); 
         }
      if (str.equals("NO RESPONSE")
          && (System.currentTimeMillis()-lastAlive)>maxWaitForClient)
	{ alive=false;
        }
      if (str.equals("KILL ME"))
	{ lastAlive= (long) 0;
          alive=false;
        }
      Thread.yield(); 
    }
  }

 public void sendUpdate(String upd)
  { tcpoutput.println(upd);
  }

 public String getUpdate()
  { String str;
    try { str=tcpinput.readLine(); }
    catch (IOException e) { errout("No response from client!"); 
                            str="NO RESPONSE"; }
    if (str==null) { str="NO RESPONSE"; } 
    return str;
  }

} // end of class ClientEntry

// -----------------------------------

class StillAliveThread extends Thread
{ StatueServer serverThread;
  void errout(String str) { System.out.println(str); }

 public StillAliveThread(StatueServer serv) 
  { serverThread=serv;
    this.start();
  }
 
 public void run()
  { while (true)
    { int i;
      try { sleep(10000); }
      catch (InterruptedException ex) { errout("StilAliveThread was interrupted!"); }
      synchronized(serverThread.clients)
       { for(i=0; i<serverThread.clients.size();i++)
          { long now=System.currentTimeMillis();
            ClientEntry cl=(ClientEntry) serverThread.clients.elementAt(i);
            if (now-cl.lastAlive>ClientEntry.maxWaitForClient) 
            { cl.disconnect();
              serverThread.clients.removeElementAt(i);
              i--;
              errout("Entry for Client " 
                     + cl.tcpsock.getInetAddress().getHostName()
                     +" removed from Server!");
            }
          }
	}
    }
  }

} // end of class StillAliveThread



